# Import libraries
from visioncam.xgradcam import XgradCAM
import keras
from keras.applications import Xception
import matplotlib.pyplot as plt

# Build Your Model
model = Xception(weights="imagenet")

# Preapare Images
image = XgradCAM.get_img_array(img_path = '/content/sample.jpg',target_size = (299,299,3) )

# Model Predcition
preds = model.predict(image)
class_idx = np.argmax(preds[0])

# Define index for your classes
class_index = {'tiger_cat': 282, 'beagle': 162 ,'african_elephant' : 386 , 'bird' : 10 }

# For All Methods
grad_cam = XgradCAM(model ,class_index['tiger_cat'] )

# For AblationCAM
last_conv_l_name = "block14_sepconv2_act"
classifier_l_names = ["avg_pool","predictions"]
grad_cam = AblationCAM(model,last_conv_layer_name = last_conv_l_name, classifier_layer_names = classifier_l_names )

preprocess_input = keras.applications.xception.preprocess_input
img_array = preprocess_input(image)
grad_heatmap = grad_cam.compute_cam_features(img_array)
plt.imshow(grad_heatmap)

XgradCAM.save_and_display_gradcam('/content/sample.jpg', grad_heatmap)


from visioncam.base import Base
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from skimage.transform import resize

class XgradCAM(Base):
    
    def __init__(self, model, class_idx, layer_name=None):
        
        Base.__init__(self , model , layer_name )
        self.class_idx=class_idx
  
        
    def compute_cam_features(self, image):
      grad_model = tf.keras.models.Model(inputs=[self.model.inputs],
                                        outputs=[self.model.get_layer(self.layer_name).output, self.model.output])
      with tf.GradientTape() as tape:
          inputs = tf.cast(image, tf.float32)
          conv_outs, predictions = grad_model(inputs)
          y_c = predictions[:, self.class_idx]

      grads = tape.gradient(y_c, conv_outs)
      weights = tf.reduce_mean(grads, axis=(0, 1, 2))
      
      X_weights = tf.reduce_sum(grads * conv_outs, axis=(0, 1, 2))
      X_weights = X_weights / (tf.reduce_sum(conv_outs, axis=(0, 1, 2)) + 1e-8)

      last_conv_layer_output = conv_outs[0]
      X_cam = tf.matmul(last_conv_layer_output, X_weights[..., tf.newaxis])

      X_cam = tf.squeeze(X_cam)

      X_cam = tf.maximum(X_cam, 0) / tf.reduce_max(X_cam)

      return X_cam.numpy()



from visioncam.base import Base
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from skimage.transform import resize

class GradCAMPlusPlus(Base):
    
    def __init__(self, model, class_idx, layer_name=None):
        
        Base.__init__(self , model , layer_name )
        self.class_idx=class_idx
        
    def compute_cam_features(self, image):

        grad_model = tf.keras.models.Model(inputs=[self.model.inputs],
                                           outputs=[self.model.get_layer(self.layer_name).output, self.model.output])
        with tf.GradientTape() as gtape1:
          with tf.GradientTape() as gtape2:
            with tf.GradientTape() as gtape3:
                img_tensor = tf.cast(image, tf.float32)
                conv_output, predictions = grad_model(img_tensor)
                output = predictions[:, self.class_idx]
                conv_first_grad = gtape3.gradient(output, conv_output)
            conv_second_grad = gtape2.gradient(conv_first_grad, conv_output)
        conv_third_grad = gtape1.gradient(conv_second_grad, conv_output)

        global_sum = np.sum(conv_output, axis=(0, 1, 2))
        alpha_num = conv_second_grad[0]
        alpha_denom = conv_second_grad[0]*2.0 + conv_third_grad[0]*global_sum
        alpha_denom = np.where(alpha_denom != 0.0, alpha_denom, 1e-10)

        alphas = alpha_num/alpha_denom
        alpha_normalization_constant = np.sum(alphas, axis=(0,1))
        alphas /= alpha_normalization_constant

        weights = np.maximum(conv_first_grad[0], 0.0)

        deep_linearization_weights = np.sum(weights*alphas, axis=(0,1))
        grad_cam_map = np.sum(deep_linearization_weights*conv_output[0], axis=2)

        heatmap = np.maximum(grad_cam_map, 0)
        max_heat = np.max(heatmap)
        if max_heat == 0:
           max_heat = 1e-10
        heatmap /= max_heat
        
        return heatmap


from visioncam.base import Base
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from skimage.transform import resize

class HriesCAM(Base):
    
    def __init__(self, model, class_idx, layer_name=None):
        
        Base.__init__(self , model , layer_name )
        self.class_idx=class_idx

        
    def compute_cam_features(self, image):

        grad_model = tf.keras.models.Model(inputs=[self.model.inputs],
                                           outputs=[self.model.get_layer(self.layer_name).output, self.model.output])
        with tf.GradientTape() as tape:
            inputs = tf.cast(image, tf.float32)
            conv_outs, predictions = grad_model(inputs)
            y_c = predictions[:, self.class_idx]

        grads = tape.gradient(y_c, conv_outs) 
        
        last_conv_layer_output = conv_outs[0]
       
        
        heatmap = last_conv_layer_output * grads
        heatmap = np.sum(heatmap, axis=-1)
        heatmap = tf.squeeze(heatmap)

        heatmap = tf.maximum(heatmap, 0) / tf.math.reduce_max(heatmap)
        
        return heatmap.numpy()


from visioncam.base import Base
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from skimage.transform import resize
import cv2

class ScoreCaM(Base):
    
    def __init__(self, model, class_idx, layer_name=None):
        
        Base.__init__(self , model , layer_name )
        self.class_idx=class_idx


    def softmax(self,x):
      f = np.exp(x)/np.sum(np.exp(x), axis = 1, keepdims = True)
      return f   

    def compute_cam_features(self, image, max_N =-1):
        img_array = tf.cast(image, tf.float32)
        cls = np.argmax(self.model.predict(img_array))

        grad_model = tf.keras.models.Model(inputs=self.model.inputs,
                                           outputs=self.model.get_layer(self.layer_name).output)

        act_map_array  = grad_model.predict(img_array)
        if max_N != -1:
            act_map_std_list = [np.std(act_map_array[0,:,:,k]) for k in range(act_map_array.shape[3])]
            unsorted_max_indices = np.argpartition(-np.array(act_map_std_list), max_N)[:max_N]
            max_N_indices = unsorted_max_indices[np.argsort(-np.array(act_map_std_list)[unsorted_max_indices])]
            act_map_array = act_map_array[:,:,:,max_N_indices]

        input_shape = self.model.layers[0].output_shape[0][1:]  
        act_map_resized_list = [cv2.resize(act_map_array[0,:,:,k], input_shape[:2], interpolation=cv2.INTER_LINEAR) for k in range(act_map_array.shape[3])]
        act_map_normalized_list = []
        for act_map_resized in act_map_resized_list:
            if np.max(act_map_resized) - np.min(act_map_resized) != 0:
                act_map_normalized = act_map_resized / (np.max(act_map_resized) - np.min(act_map_resized))
            else:
                act_map_normalized = act_map_resized
            act_map_normalized_list.append(act_map_normalized)
        masked_input_list = []
        for act_map_normalized in act_map_normalized_list:
            masked_input = np.copy(img_array)
            for k in range(3):
                masked_input[0,:,:,k] *= act_map_normalized
            masked_input_list.append(masked_input)
        masked_input_array = np.concatenate(masked_input_list, axis=0)
        pred_from_masked_input_array = self.softmax(self.model.predict(masked_input_array))
        weights = pred_from_masked_input_array[:,cls]
        cam = np.dot(act_map_array[0,:,:,:], weights)
        cam = np.maximum(0, cam)  
        cam /= np.max(cam)  

        
        return cam


from visioncam.base import Base
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from skimage.transform import resize

class GradCAM(Base):
    
    def __init__(self, model, class_idx, layer_name=None):
        
        Base.__init__(self , model , layer_name )

        self.class_idx=class_idx
        
        
    def compute_cam_features(self, image):

        grad_model = tf.keras.models.Model(inputs=[self.model.inputs],
                                           outputs=[self.model.get_layer(self.layer_name).output, self.model.output])
        with tf.GradientTape() as tape:
            inputs = tf.cast(image, tf.float32)
            conv_outs, predictions = grad_model(inputs)
            y_c = predictions[:, self.class_idx]

        grads = tape.gradient(y_c, conv_outs) 
        pooled_grads = tf.reduce_mean(grads, axis=(0, 1, 2))

        last_conv_layer_output = conv_outs[0]
        heatmap = last_conv_layer_output @ pooled_grads[..., tf.newaxis]
        heatmap = tf.squeeze(heatmap)

        heatmap = tf.maximum(heatmap, 0) / tf.math.reduce_max(heatmap)
        
        return heatmap.numpy()


import numpy as np
import tensorflow as tf
from keras.applications.xception import preprocess_input
from tensorflow.keras.preprocessing.image import load_img,img_to_array,array_to_img
from tensorflow import keras
from IPython.display import Image
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from visioncam.base import Base

class AblationCAM(Base):
    def __init__(self, model,last_conv_layer_name,classifier_layer_names):
        
        self.model = model
        self.last_conv_layer_name=last_conv_layer_name
        self.classifier_layer_names=classifier_layer_names
        
    def get_image(image_path,img_size):
        orig = load_img(image_path, target_size=img_size)
        image = img_to_array(orig)
        image = np.expand_dims(image, axis=0)
        img_array =preprocess_input(image)
        return img_array

    def compute_cam_features(self, img):
        preds = self.model.predict(img)
        last_conv_layer = self.model.get_layer(self.last_conv_layer_name)
        last_conv_layer_model = keras.Model(self.model.inputs, last_conv_layer.output)
        classifier_input = keras.Input(shape=last_conv_layer.output.shape[1:])
        x = classifier_input
        for layer_name in self.classifier_layer_names:
            x = self.model.get_layer(layer_name)(x)
        classifier_model = keras.Model(classifier_input, x)
        activations = last_conv_layer_model.predict(img) 
        last_layer_activation = activations[-1]

        

        last_layer_activation = np.expand_dims(last_layer_activation,axis=0)
        heatmap = np.mean(last_layer_activation, axis=-1)
        heatmap = np.maximum(heatmap, 0) / np.max(heatmap)
        values = []
        
        for i in range(last_layer_activation.shape[-1]):
          activations = last_conv_layer_model.predict(img) 
          last_layer_activation = activations[-1]
          activation = last_layer_activation
          activation = np.expand_dims(activation,axis=0)
          jj = int(last_layer_activation.shape[0])
          kk = int(last_layer_activation.shape[1])
          
          for j in range(jj):
            for k in range(kk):
              activation[0][j][k][i]=0.
          prediction = classifier_model(activation)
          a= prediction[0][404]
          values.append(a.numpy())

        aero = 0.8209318
        aero = [aero]*last_layer_activation.shape[-1]
        weight_ratio = []
        for i in range(last_layer_activation.shape[-1]):
          b= (float(aero[i]) - float(values[i]))/float(aero[i])
          weight_ratio.append(b)
        last_layer_activation = np.expand_dims(last_layer_activation,axis=0)
        for i in range(last_layer_activation.shape[-1]):
          for  j in range(last_layer_activation.shape[1]):
            for k in range(last_layer_activation.shape[2]):
              last_layer_activation[0][j][k][i]=last_layer_activation[0][j][k][i]*weight_ratio[i]

        heatmap = np.mean(last_layer_activation, axis=-1)
        heatmap = np.maximum(heatmap, 0) / np.max(heatmap)
        heatmap=heatmap[0,:,:,]
        
        return heatmap



import tensorflow as tf
from tensorflow.keras.applications import imagenet_utils
import numpy as np
import matplotlib.pyplot as plt
from skimage.transform import resize
import keras
from IPython.display import Image, display
import matplotlib as mpl
import keras


class Base:
    def __init__(self, model, layer_name=None):
        self.model = model
        self.layer_name = layer_name

        
        if not layer_name:
            self.layer_name = self.find_target_layer()

    def find_target_layer(self):
        for layer in reversed(self.model.layers):
            if len(layer.output_shape) == 4:
                return layer.name
        raise ValueError('Could not find 4D layer. Cannot apply CAM .')
    
    @staticmethod
    def get_img_array(img_path, target_size = (299,299,3)):
        img = keras.utils.load_img(img_path, target_size=target_size)
        array = keras.utils.img_to_array(img)
        array = np.expand_dims(array, axis=0)
        return array

    def compute_cam_features(self, image):
        cam_model = tf.keras.models.Model(inputs=[self.model.inputs],
                                          outputs=[self.model.get_layer(self.layer_name).output, self.model.output])
        features, outputs = cam_model.predict(image)
        global_avg_pool_weights = self.model.layers[-1].get_weights()[0]
        image_feature = features[0]
        cam_features = resize(image_feature, (224, 224))
        pred = np.argmax(outputs[0])
        cam_weights = global_avg_pool_weights[:, pred]
        heatmap = np.dot(cam_features, cam_weights)
        return heatmap

    
    @staticmethod
    def save_and_display_gradcam(img_path, heatmap, cam_path="cam.jpg", alpha=0.4):
        img = keras.utils.load_img(img_path)
        img = keras.utils.img_to_array(img)

        heatmap = np.uint8(255 * heatmap)

        jet = mpl.colormaps["jet"]

        jet_colors = jet(np.arange(256))[:, :3]
        jet_heatmap = jet_colors[heatmap]

        jet_heatmap = keras.utils.array_to_img(jet_heatmap)
        jet_heatmap = jet_heatmap.resize((img.shape[1], img.shape[0]))
        jet_heatmap = keras.utils.img_to_array(jet_heatmap)

        superimposed_img = jet_heatmap * alpha + img
        superimposed_img = keras.utils.array_to_img(superimposed_img)

        superimposed_img.save(cam_path)

        display(Image(cam_path))



import setuptools


setuptools.setup(
    name = "visioncam",
    version = "0.0.1",
    license='MIT',
    author = 'Ahmed Tolba',
    author_email = 'atolba210@gmail.com',
    description = 'XAI FrameWork for computer vision tasks',
    packages = setuptools.find_packages(),

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
   
   

    )


